#!/usr/bin/python
"""
Command line script for interacting with either the remote webservice, or a local Plasmo Model Repository.
Default mode is to run remotely, i.e. using the webflow website.

local models are stored like this:
$REPO_PATH/models/acm/acm_v1/acm_v1.sml

remote models are stored like this:
http://webflow.simileweb.com/models/acm/acm_v1/acm_v1.sml

To make the script usable from any directory, add its containing directory to system path.
(e.g. linux: 

.. code-block:: bash 

    $ export PATH=$PATH:/path/to/directory
)


Example usage:
++++++++++++++

Run one-off conversion 'sml_to_pl'. This formats a request to the webservice, uploads the file (acm_v1.sml), and outputs the result to acm_v1.pl:

.. code-block:: bash 

    $ python webflow_client.py -d sml_to_pl -f $REPO_PATH/models/acm/acm_v1/acm_v1.sml -o acm_v1.pl

Run with settings 'remote_test', runs through a couple of conversions for some models:
(see webflow_client_settings.py for examples of settings)

.. code-block:: bash 

    $ python webflow_client.py -s remote_test

Run workflow locally with steps 'sml_to_pl', 'pl_to_xml', on model 'lintul' (defaults to version v1):

.. code-block:: bash 

    $ python webflow_client.py -l -w ['sml_to_pl', 'pl_to_xml'] -m ['lintul']

Command line options:
+++++++++++++++++++++
"""
# Note, I add the command line help options to the doc string: hence title with no content above.

import os, re, sys
import optparse
from  urllib2 import urlopen, Request
from xml.etree import ElementTree as et

from poster.encode import multipart_encode
from poster.streaminghttp import register_openers

import webflow_client_settings

# Setup parser here so as I can add it to __doc__
def parse_options():
    parser = setup_options()
    return parser.parse_args()

def setup_options():
    parser = optparse.OptionParser()
    parser.add_option("-s", "--settings", dest="settings", help="Name of settings to use. Overrides all other options.")

    parser.add_option("-l", "--local", action="store_true", dest="local", help="run on local repository")
    parser.add_option("-u", "--url", dest="url", help="URL of webservice")
    parser.add_option("-p", "--path", dest="path", help="Path of local Plasmo Model repository")

    parser.add_option("-w", "--workflow", dest="workflow",  help="List of processes in format ['p1', 'p2'].")
    parser.add_option("-m", "--models",   dest="models",    help="List of models in format ['m1', 'm2'].")

    parser.add_option("-d", "--do-process", dest="process",  help="Name of process")
    parser.add_option("-f", "--infile",  dest="in_file",  help="Name of file")
    parser.add_option("-o", "--outfile", dest="out_file", help="Name of file to write results to")

    parser.add_option("-v", "--verbose", action="store_true", dest="verbose", help="print mini reports")
    parser.add_option("-t", "--test", action="store_true", dest="test", help="run through test")
    parser.set_usage('./webflow_client.py [options]')

    return parser

# Add parser.format_help() (i.e. help string you get if you do $ ./webflow_client.py -h)
__doc__ += setup_options().format_help(optparse.IndentedHelpFormatter(2, 60, 150))

PLASMO_MODELS_PATH = os.environ['PLASMO_MODELS_PATH']

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'

    def disable(self):
        self.HEADER = ''
        self.OKBLUE = ''
        self.OKGREEN = ''
        self.WARNING = ''
        self.FAIL = ''
        self.ENDC = ''


def traverse_model_dirs(callback):
    """
    Function used to traverse all the models in the plasmo model repo
    """
    # TODO: replace using below classes.
    reports = []

    # get to the right place.
    curr_dir = os.getcwd()
    os.chdir(os.path.join(PLASMO_MODELS_PATH))
    models = os.listdir(".")
    models.sort()
    # models are stored like this:
    # $REPO_PATH/models/acm/acm_v1/acm_v1.sml
    # where there can be any number of versions under the main model file
    for model in models:
	if re.match("\.hg", model):
	    continue
        #if model in skip:
            #continue
        found_sml = False
        print("%s%s%s" % (bcolors.HEADER, model, bcolors.ENDC))
        # e.g. $REPO_PATH/models/acm/
        os.chdir(os.path.join(PLASMO_MODELS_PATH,  model))
        model_versions = os.listdir(".")
        model_versions.sort()
        for model_version in model_versions:
            if re.match("%s_" % model, model_version) and os.path.isdir(model_version):
                # e.g. $REPO_PATH/models/acm/acm_v1/
                os.chdir(model_version)
                print " %s" % model_version
            else:
                continue

            file_names = os.listdir(".")
            #reports, found_sml = process_version(file_names, reports)
            reports, found_sml = callback(model, model_version, file_names, reports)

        if not found_sml:
            report = [model]
            reports.append(report)

    # e.g. $REPO_PATH/models/acm/
    os.chdir(os.path.join(PLASMO_MODELS_PATH, model))

    os.chdir(curr_dir)
    return reports

def construct_url(*args):
    """helper function to build url from args"""
    # TODO: should be smarter and check '/'s are in right place.
    url = "".join(args)
    #print url
    return url

def recursive_remove(model_dir):
    """function to delete a dir and all subdirs/files"""
    dirs = []
    for dir, dirs, files in os.walk(model_dir):
	#print dir, dirs, files

	for file in files:
	    #print 'delete file %s' % file
	    os.remove(os.path.join(dir, file))

	try:
	    os.removedirs(dir)
	except:
	    pass

# Use this for all classes?
class WeblfowBase(object):
    def __init__(self, **kwargs):
	for k, v in kwargs:
	    setattr(self, k, v)

class ModelList(list):
    """List class for holding all models"""
    def __init__(self, type, base_location="http://webflow.simileweb.com/"):
	self.base_location = base_location

	self.get_models = getattr(self, '%s_get_models' % type)
	models = self.get_models()
	super(ModelList, self).__init__(models)

    def local_get_models(self):
	abs_path = os.path.join(self.base_location, 'models')
	model_dirs = os.listdir(abs_path)
	model_dirs.sort()
	models = [Model('local', self.base_location, model_dir) for model_dir in model_dirs]
	return models

    def remote_get_models(self):
	url = self.base_location
	models = []
        f = urlopen(construct_url(url, 'models/'))
        html = f.read()
	doc = et.fromstring(html)
        table = doc.find("body/div/div/div/table")
        for row in table.findall('tr'):
	    for td in row.findall('td'):
		try:
		    a = td.find('a')
		    if a.attrib['class'] != 'model':
			raise
		    models.append(Model('remote', url, a.text))
		except Exception as e:
		    #print e
		    pass
	return models


class Model(object):
    def __init__(self, type, base_location, name, create=False):
	self.name = name
	self.base_location = base_location
	self.get_versions = getattr(self, '%s_get_versions' % type)
	self.create = getattr(self, '%s_create' % type)
	self.delete = getattr(self, '%s_delete' % type)
	if create:
	    self.create()
	self.versions = self.get_versions(base_location)

    def local_delete(self):
	# Deletes all files and dirs under dir it gets passed. User beware.
	recursive_remove(os.path.join(self.base_location, self.name))

    def remote_delete(self):
	f = urlopen(construct_url(self.base_location, 'models/'), data='model_name=%s&_method=delete' % self.name)


    def local_create(self):
	try:
	    os.mkdir(os.path.join(self.base_location, self.name))
	    os.mkdir(os.path.join(self.base_location, self.name, "%s_v1" % self.name))
	except:
	    pass

    def remote_create(self):
	f = urlopen(construct_url(self.base_location, 'models/'), data='name=%s' % self.name)


    def local_get_versions(self, base_path):
	versions = []
	abs_path = os.path.join(base_path, self.name)
	dirs = os.listdir(abs_path)
	dirs.sort()
	for dir in dirs:
	    if re.search("[\w\-_]_v\d", dir):
		versions.append(Version('local', base_path, self.name, dir))
	return versions

    def remote_get_versions(self, url):
	versions = []
        f = urlopen(construct_url(url, 'models/', self.name, '/'))
        html = f.read()
	doc = et.fromstring(html)
        table = doc.find("body/div/div/div/table")
        for row in table.findall('tr'):
	    for td in row.findall('td'):
		a = td.find('a')
		versions.append(Version('remote', url, self.name,  a.text))
	return versions


class Version(object):
    def __init__(self, type, base_location, model_name, name):
	verbose("Creating version %s" % name)
	self.type = type
	self.base_location = base_location
	self.model_name = model_name
	self.name = name

	# set methods depending on whether local or remote (i.e. on type)
	self.get_files = getattr(self, '%s_get_files' % type)
	self.add_file = getattr(self, '%s_add_file' % type)
	self.process_file = getattr(self, '%s_process_file' % type)

	self.get_files(base_location)
	#self.files = self.get_files(base_location)
	#self.steps = self.get_steps(base_location)

    def local_delete(self):
	# TODO:
	pass

    def remote_delete(self):
	data = 'model_name=%s&version_name=%s&_method=delete' % (self.model_name, self.name)
	f = urlopen(construct_url(self.base_location, 'models/', self.model_name, '/'), data=data)


    def local_add_file(self, file_path):
	#print 'basename: ', os.path.basename(file_path)
	out_path = os.path.join(self.base_location, self.model_name, self.name, os.path.basename(file_path))
	print 'add file: ', out_path
	open(out_path, 'w').write(open(file_path).read())

    def remote_add_file(self, file_path):
	try:
	    file = open(file_path, 'r')
	except:
	    # TODO: error.
	    pass
	# I don't know how this works. This is magic. It pops up as a regular file in django though.
	register_openers()

	# TODO: wont work for .gen.pl
	ext = os.path.splitext(file_path)[1][1:]
	datagen, headers = multipart_encode({ext: file, '_method': 'add-files'})

	request = Request(construct_url('http://webflow.simileweb.com/', 'models/', self.model_name, '/', self.name, '/'), datagen, headers)
	f = urlopen(request)


    # These could have gone either here or in the File class. It's here for consistency
    # with Django views VersionDispatch.
    def remote_process_file(self, in_file_name, process, out_file_pre_ext=''):
	out_file_ext = process['out_file_ext']
	print out_file_ext
	out_file_name = "%s%s%s" % (os.path.splitext(in_file_name)[0], out_file_pre_ext, out_file_ext)

	data = "in_file_name=%s&out_file_name=%s&process_name=%s" % (in_file_name, out_file_name, process['name'])

	print 'about to process file'
        f = urlopen(construct_url(self.base_location, 'models/', self.model_name, '/', self.name, '/'), data=data)
	#print f.read()

    def local_process_file(self, in_file_name, process, out_file_pre_ext=''):
	print 'about to process file'

	in_file = self.get_file(in_file_name)
	output = process.do_process(in_file.path)
	out_file_name = "%s%s%s" % (os.path.splitext(in_file_name)[0], out_file_pre_ext, process.out_filetype())

	if not self.get_file(out_file_name):
	    out_file = File(self.type, self.base_location, self, out_file_name, content=output)
	    self.files.append(out_file)
	else:
	    out_file = File(self.type, self.base_location, self, out_file_name, content=output)

    def get_file(self, file_name):
	found = False
	for file in self.files:
	    if file.name == file_name:
		found = True
		break

	if found:
	    return file
	else:
	    return None

    def local_get_files(self, base_path):
	abs_path = os.path.join(base_path, self.model_name, self.name)
	file_names = os.listdir(abs_path)
	file_names.sort()
	files = [File('local', base_path, self, file_name, os.path.join(abs_path, file_name)) for file_name in file_names]
	self.files = files

    def remote_get_files(self, url):
	files = []
        f = urlopen(construct_url(url, 'models/', self.model_name, '/', self.name, '/'))
        html = f.read()
	self.doc = et.fromstring(html)
	for table in self.doc.findall("body/div/div/div/table"):
	    try:
		if table.attrib['id'] == 'files':
		    # got the table i want.
		    break 
	    except:
		pass

        for row in table.findall('tr'):
	    for td in row.findall('td'):
		a = td.find('a')
		if a.attrib['class'] == 'file':
		    files.append(File('remote', url, self, a.text, a.attrib['href']))
	self.files = files


class File(object):
    def __init__(self, type, base_path, version, name, path='', content=''):
	self.type = type
	self.base_path = base_path
	self.version = version
	self.name = name
	self.path = path
	self.content = content
	if self.content:
	    if type == 'local':
		self.path = os.path.join(self.base_path, self.version.model_name, self.version.name, self.name)
		open(self.path, 'w').write(self.content)
	    else:
		#self.
		pass

    def get_content(self):
	if self.content != '':
	    return self.content
	else:
	    return getattr(self, "%s_get_content" % self.type)()

    def local_get_content(self):
	self.content = open(self.path, 'r').read()
	return self.content

    def remote_get_content(self):
	f = urlopen(construct_url(self.base_path, self.path))
	self.content = f.read()
	return self.content

    def __repr__(self):
	return "<File %s>" % self.name



def test(options):
    # TODO: broken
    print "Running tests %s" % ('remotely' if not options.local else 'locally')
    print "Getting all Models, Versions and Files"
    if options.local:
	model_list = ModelList('local', PLASMOPATH)
    else:
	# Default.
	if options.url:
	    model_list = ModelList('remote', options.url)
	else:
	    model_list = ModelList('remote')

    for model in model_list:
	print "  ", model.name
	for version in model.versions:
	    print "    ", version.name
	    for file in version.files:
		print "      ", file.name
		file.get_content()
    return 0


def process_file(local, base_location, file_path, process_name):
    try:
	file = open(file_path, 'r')
    except:
	# TODO: error.
	sys.stderr.write('Couldn\'t open file %s' % (file_path))
	exit(1)

    if local:
	process = local_get_process(process_name)
	output = process.do_process(file_path)
	return output
    else:
	register_openers()
	datagen, headers = multipart_encode({'in_file': file})
	request = Request(construct_url(base_location, 'processes/', process_name, '/'), datagen, headers)
	f = urlopen(request)
	return f.read()


def get_options(options):
    parser = setup_options()
    try:
	settings = getattr(webflow_client_settings, options.settings)
    except AttributeError:
	settings_names = []
	for name in dir(webflow_client_settings):
	    if re.match("^__[\w_]+__$", name) or name == 'Default':
		continue
	    settings_names.append(name)

	sys.stderr.write('Settings name not recognised, choices are: \n')
	sys.stderr.write(" * %s\n" % ("\n * ".join(settings_names)))
	exit(1)

    # if settings are provided, they completely override options:
    # Warn user if they have provided any other options.
    for option in parser.option_list:
	if option.dest and (option.dest != 'settings') and (option.dest != 'verbose'):
	    if getattr(options, option.dest):
		print "Warning: option --%s will be overridden." % option.dest

    return settings

def local_setup_location(options):
    if options.path:
	plasmo_path = options.path
    else:
	try:
	    plasmo_path = os.environ['PLASMOPATH']
	except: 
	    sys.stderr.write('environmental variable PLASMOPATH must be set, or use --path option.\n')
	    exit(1)
    if not os.path.isdir(plasmo_path):
	sys.stderr.write("Plasmo model directory %s does not exist.\n" % plasmo_path)
	exit(1)

    # TODO: HACK!
    #plasmo_path = '/home/markmuetz/projects/plasmo-mm'
    return plasmo_path

def local_setup_processes():
    # Load all process modules into a dict for (possibly?) faster access.
    process_modules = {}
    for file_name in os.listdir('processes'):
	process_name, ext = os.path.splitext(file_name)
	if ext != '.py' or process_name == '__init__':
	    continue
	process_modules[process_name] = local_get_process(process_name)
    return process_modules

def local_get_process(process_name):
    try:
	__import__('webflow_tools.processes.%s' % (process_name.replace('-', '_')))
	module = sys.modules["webflow_tools.processes.%s" % (process_name.replace('-', '_'))]
	return module
    except Exception as e:
	sys.stderr.write('Error with process: %s\n', process_name)
	sys.stderr.write(e)
	sys.stderr.write('\n')
	exit(1)
	

def remote_setup_location(options):
    # TODO: urls
    # Note url must end in '/'
    if not options.url:
	return 'http://webflow.simileweb.com/'
    else:
	if re.search('^http:\/\/.*\/$', options.url):
	    return options.url
	else:
	    sys.stderr.write('URL must start with http and end in a slash, e.g. "http://webflow.simileweb.com/"\n')
	    exit(0)

def remote_setup_processes(url):
    f = urlopen(construct_url(url, 'processes/'))
    html = f.read()
    processes_doc = et.fromstring(html)
    list = processes_doc.find("body/div/div/div/ul")
    remote_processes = {}
    for element in list.findall('li'):
	process_name = element.find('a').text
	remote_processes[process_name] = remote_get_process(url, process_name)
    return remote_processes

def remote_get_process(url, process_name):
    # TODO: make these look like local process modules?
    # if only i hadn't defined position et al as funcs, would be easy...
    #class Process:
	#pass
    #p = Process()

    f = urlopen(construct_url(url, 'processes/', process_name, '/'))
    html = f.read()
    process_doc = et.fromstring(html)
    table = process_doc.find("body/div/div/div/table")
    remote_process = {}
    for row in table.findall('tr'):
	remote_process[row.attrib['id'].replace('-', '_')] = row.findall('td')[1].text
    return remote_process

def do_workflow(local, base_location, workflow, models, processes):
    if local:
	local_remote = 'local'
    else:
	local_remote = 'remote'

    if models == 'all':
	model_list = ModelList(local_remote, base_location)
    else:
	model_list = [Model(local_remote, base_location, model_name) for model_name in models]

    for model in model_list:
	print "  ", model.name
	for version in model.versions:
	    print "    ", version.name
	    for step in workflow:
		process = processes[step['name']]

		try:    in_file_pre_ext = step['in_file_pre_ext']
		except: in_file_pre_ext = ''

		try:    out_file_pre_ext = step['out_file_pre_ext']
		except: out_file_pre_ext = ''

		if local:
		    in_file_name  = "%s%s%s" % (version.name, in_file_pre_ext, process.in_filetype())
		    #out_file_name = "%s%s%s" % (version.name, out_file_pre_ext, process.out_filetype())
		else:
		    in_file_name  = "%s%s%s" % (version.name, in_file_pre_ext, process['in_file_ext'])
		    #out_file_name = "%s%s%s" % (version.name, out_file_pre_ext, process['out_file_ext'])
		version.process_file(in_file_name, process, out_file_pre_ext)


def main(options, args):
    if options.verbose:
	def verbose(string):
	    """print if verbose mode is set"""
	    print(string)
    else:
	def verbose(string):
	    pass
    # naughty!
    globals()['verbose'] = verbose



    if options.settings:
	# Override all other options. Warn user if they've set any.
	options = get_options(options)

    if options.local:
	# base_location is a path. processes is a dict of process modules.
	base_location = local_setup_location(options)
    else:
	# base_location is a URL. processes is a dict of online processes.
	base_location = remote_setup_location(options)

    if options.workflow:
	# If these are options evaluate the workflow/models strings
	# If from settings should already be lists (but check anyway)
	if type(options.workflow) is str:
	    options.workflow = eval(options.workflow)
	if not type(options.workflow) is list:
	    sys.stderr.write('Workflow must be in form ["p1", "p2"] where p1 etc are process names.\n')
	    exit(1)

	if type(options.models) is str:
	    options.models = eval(options.models)
	if not type(options.models) is list:
	    sys.stderr.write('Models must be in form ["m1", "m2"] where m1 etc are model names.\n')
	    exit(1)

	if options.local:
	    processes     = local_setup_processes()
	else:
	    processes     = remote_setup_processes(base_location)

	do_workflow(options.local, base_location, options.workflow, options.models, processes)
	exit(0)

    if options.process:
	if not options.in_file:
	    sys.stderr.write('You must provide an in file (-f or --infile)\n')
	    exit(1)
	#if options.local:
	    #process = local_get_process(options.process)
	#else:
	    #process = remote_get_process(options.process)

	output = process_file(options.local, base_location, options.in_file, options.process)

	if options.out_file:
	    open(options.out_file, 'w').write(output)
	else:
	    print output
	exit(0)

    if options.test:
	# put client through its paces.
	exit(test(options))

    sys.stderr.write('Problem parsing options.\n\n')
    parser = setup_options()
    parser.print_help()
    exit(1)

if __name__ == "__main__":
    options, args = parse_options()
    main(options, args)
